"""add_shared_memory_to_resource_presets

Revision ID: c481d3dc6c7d
Revises: f5530eccf202
Create Date: 2020-04-20 14:10:35.591063

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = 'c481d3dc6c7d'
down_revision = 'f5530eccf202'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column('resource_presets', sa.Column('shared_memory', sa.BigInteger(), nullable=True))
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column('resource_presets', 'shared_memory')
    # ### end Alembic commands ###
